KeystoneManager = LibStub('AceAddon-3.0'):NewAddon('KeystoneManager', 'AceConsole-3.0', 'AceEvent-3.0', 'AceTimer-3.0');
---@type StdUi
local StdUi = LibStub('StdUi');

local icon = LibStub('LibDBIcon-1.0');
local ldb = LibStub:GetLibrary('LibDataBroker-1.1');

local shortNames = {
	[197] = 'EOA',
	[198] = 'DHT',
	[199] = 'BRH',
	[200] = 'HOV',
	[206] = 'NL',
	[207] = 'VOTW',
	[208] = 'MOS',
	[209] = 'ARC',
	[210] = 'COS',
	[233] = 'COEN',
	[227] = 'KARA:L',
	[234] = 'KARA:U',
	[239] = 'SOTT',

	[244] = 'AD',
	[245] = 'FH',
	[246] = 'TD',
	[247] = 'ML',
	[248] = 'WM',
	[249] = 'KR',
	[250] = 'TOS',
	[251] = 'UNDR',
	[252] = 'SOTS',
	[353] = 'SIEGE',
}


local activities = {
	[197] = 459,
	[198] = 460,
	[199] = 463,
	[200] = 461,
	[206] = 462,
	[207] = 464,
	[208] = 465,
	[209] = 467,
	[210] = 466,
	[233] = 476,
	[227] = 471,
	[234] = 473,
	[239] = 486,


	[244] = 502,
	[245] = 518,
	[246] = 526,
	[247] = 510,
	[248] = 530,
	[249] = 514,
	[250] = 504,
	[251] = 507,
	[252] = 522,
	[353] = 534,
}

local defaults = {
	enabled = true,
	announce = true,
	excludes = {},
	keystones = {},
	target = 'GUILD',
	whisper = '',
	minLevel = 0,
	maxLevel = 20,
	ldbStorage = {
		hide = false
	}
};

local kmldbObject = {
	type = 'launcher',
	text = 'Keystone Manager',
	label = 'Keystone Manager',
	icon = 'Interface\\Icons\\INV_Relics_Hourglass',
	OnClick = function() KeystoneManager:ShowWindow(); end,
	OnTooltipShow = function(tooltip)

		local info = KeystoneManager:GetCurrentKeystoneInfo();
		local name = KeystoneManager:NameAndRealm();

		if info then
			tooltip:AddLine('Current:');
			tooltip:AddDoubleLine(
				format('|cffffffff%s|r', KeystoneManager:NameWithoutRealm(name)),
				KeystoneManager:FormatKeystone(info)
			);
			tooltip:AddLine(' ');
		end


		tooltip:AddLine('Other Keys:');
		for char, key in pairs(KeystoneManager.db.keystones) do
			if name ~= char then
				tooltip:AddDoubleLine(
					format('|cffffffff%s|r', KeystoneManager:NameWithoutRealm(char)),
					KeystoneManager:FormatKeystone(key)
				);
			end
		end
	end,
};

function KeystoneManager:RegisterOptionWindow()
	if self.optionsFrame then
		return;
	end
	
	self.optionsFrame = StdUi:PanelWithTitle(UIParent, 100, 100, 'Keystone Manager');
	self.optionsFrame.name = 'Keystone Manager';
	self.optionsFrame:Hide();

	local enabled = StdUi:Checkbox(self.optionsFrame, 'Enable Addon');
	if self.db.enabled then enabled:SetChecked(true); end
	enabled.OnValueChanged = function(_, flag) KeystoneManager.db.enabled = flag; end
	
	local announce = StdUi:Checkbox(self.optionsFrame, 'Announce new key in party channel');
	if self.db.announce then announce:SetChecked(true);	end
	announce.OnValueChanged = function(_, flag)	KeystoneManager.db.announce = flag;	end


	StdUi:GlueTop(enabled, self.optionsFrame, 10, -40, 'LEFT');
	StdUi:GlueBelow(announce, enabled, 0, -10, 'LEFT');

	InterfaceOptions_AddCategory(self.optionsFrame);
end

function KeystoneManager:OnInitialize()
	if not KeystoneManagerDb or type(KeystoneManagerDb) ~= 'table' or KeystoneManagerDb.global then
		KeystoneManagerDb = defaults;
	end

	self.db = KeystoneManagerDb;
	
	self:RegisterOptionWindow();

	self:RegisterChatCommand('keystonemanager', 'ShowWindow');
	self:RegisterChatCommand('keylist', 'ShowWindow');
	self:RegisterChatCommand('keyprint', 'PrintKeystone');

	self:RegisterEvent('PLAYER_ENTERING_WORLD');

	self:RegisterEvent('CHALLENGE_MODE_MAPS_UPDATE', 'UpdateWeeklyBest');
	self:RegisterEvent('CHALLENGE_MODE_MEMBER_INFO_UPDATED', 'UpdateWeeklyBest');
	self:RegisterEvent('CHALLENGE_MODE_LEADERS_UPDATE', 'UpdateWeeklyBest');
	self:RegisterEvent('CHALLENGE_MODE_COMPLETED', 'UpdateWeeklyBestAndKeystone');

	if not ldb:GetDataObjectByName('KeystoneManager') then
		ldb:NewDataObject('KeystoneManager', kmldbObject);
		icon:Register('KeystoneManager', kmldbObject, self.db.ldbStorage);
	end
end

-- Can't really request it faster
function KeystoneManager:PLAYER_ENTERING_WORLD()
	if self.onceRequested then
		self:RefreshDataText();
		self:GetKeystone();
		return;
	end

	C_Timer.After(5, function()
		C_MythicPlus.RequestMapInfo();
		self:GetMapInfo();
		for mapId, mapName in pairs(self.mapNames) do
			C_ChallengeMode.RequestLeaders(mapId);
		end

		self:RefreshDataText();
		self:GetKeystone();

		self.onceRequested = true;
	end);
end

function KeystoneManager:GetMapInfo()
	self.mapNames = {};
	local maps = C_ChallengeMode.GetMapTable();
	
	for i = 1, #maps do
		local name, id = C_ChallengeMode.GetMapUIInfo(maps[i]);
		self.mapNames[id] = name;
	end
end

function KeystoneManager:UpdateWeeklyBestAndKeystone()
	C_Timer.After(2, function()
		KeystoneManager:UpdateWeeklyBest();
		KeystoneManager:GetKeystone();
	end);
end

function KeystoneManager:ShowWindow(input)
	if not self.KeystoneWindow then

		self.KeystoneWindow = StdUi:Window(UIParent, 'Keystone Manager', 625, 550);

		local window = self.KeystoneWindow;
		window.titlePanel:SetBackdrop(nil);
		window:SetPoint('CENTER');
		window:SetFrameStrata('DIALOG');

		local whisperOpts = {
			{text = 'Whisper', value = 'WHISPER'},
			{text = 'Guild', value = 'GUILD'},
			{text = 'Party', value = 'PARTY'},
			{text = 'Raid', value = 'RAID'},
			{text = 'Instance', value = 'INSTANCE_CHAT'},
		}
		local target = StdUi:Dropdown(window, 120, 24, whisperOpts, self.db.target);
		StdUi:AddLabel(window, target, 'Send Message To', 'TOP');
		StdUi:GlueTop(target, window, 10, -50, 'LEFT');
		target.OnValueChanged = function(_, value, text)
			self.db.target = value;
		end;

		local whisper = StdUi:EditBox(window, 120, 24, self.db.whisper);
		StdUi:AddLabel(window, whisper, 'Whisper target', 'TOP');
		StdUi:GlueRight(whisper, target, 10, 0);
		whisper.OnValueChanged = function(_, text)
			self.db.whisper = text;
		end;

		local btn = StdUi:Button(window, 120, 24, 'Report');
		StdUi:GlueRight(btn, whisper, 10, 0);
		btn:SetScript('OnClick', function()
			self:ReportKeys();
		end);

		local minLevel = StdUi:NumericBox(window, 120, 24, self.db.minLevel);
		StdUi:AddLabel(window, minLevel, 'Min Level', 'TOP');
		StdUi:GlueBelow(minLevel, target, 0, -30);
		minLevel:SetMinMaxValue(0, 50);
		minLevel.OnValueChanged = function(_, val)
			self.db.minLevel = val;
		end;

		local maxLevel = StdUi:NumericBox(window, 120, 24, self.db.maxLevel);
		StdUi:AddLabel(window, maxLevel, 'Max Level', 'TOP');
		StdUi:GlueRight(maxLevel, minLevel, 10, 0);
		maxLevel:SetMinMaxValue(0, 50);
		maxLevel.OnValueChanged = function(_, val)
			self.db.maxLevel = val;
		end;

		local cols = {
			{
				name  = 'Character',
				width = 120,
				index = 'name'
			},

			{
				name  = 'Weekly Best',
				width = 100,
				index = 'weeklyBest'
			},

			{
				name  = 'Dungeon',
				width = 165,
				index = 'mapName',
			},

			{
				name  = 'Level',
				width = 80,
				index = 'level',
				type  = 'number'
			},
		}
		self.ScrollTable = StdUi:ScrollTable(window, cols, 16, 20);
		StdUi:GlueAcross(self.ScrollTable, window, 10, -160, -10, 60);
		self:UpdateTable();

		self.ScrollTable:RegisterEvents({
			OnClick = function(table, cellFrame, rowFrame, rowData, columnData, rowIndex)
				local link = KeystoneManager:CreateLink(rowData);

				if link then
					if IsAltKeyDown() then
						KeystoneManager:CreateGroup(rowData);
					elseif IsShiftKeyDown() then
						HandleModifiedItemClick(link);
					else
						GameTooltip:SetOwner(UIParent);
						GameTooltip:SetHyperlink(link);
						GameTooltip:Show();
					end
				end
				return true;
			end,
		});

		-- Clear button
		local clearBtn = StdUi:Button(window, 120, 24, 'Report');
		StdUi:GlueBottom(clearBtn, window, -10, 10, 'RIGHT');
		clearBtn:SetText('Clear');
		clearBtn:SetScript('OnClick', function()
			self:ClearKeystones();
		end);

		-- Refresh button
		local refreshBtn = StdUi:Button(window, 120, 24, 'Refresh');
		StdUi:GlueBottom(refreshBtn, window, 10, 10, 'LEFT');
		refreshBtn:SetScript('OnClick', function()
			self:GetKeystone(true);
			self:UpdateWeeklyBest();
		end);

		-- Copy button
		local copyBtn = StdUi:Button(window, 120, 24, 'Copy Keys');
		StdUi:GlueRight(copyBtn, refreshBtn, 10, 0);
		copyBtn:SetScript('OnClick', function()
			self:ShowCopyWindow();
		end);
	end

	self.KeystoneWindow:Show();
end

function KeystoneManager:CreateLink(data)
	-- "|cffa335ee|Hkeystone:158923:244:2:10:0:0:0|h[Keystone: Atal'dazar (2)]|h|r"
	local link = string.format(
		'|cffa335ee|Hkeystone:158923:%d:%d:10:0:0:0|h[Keystone: %s (%d)]|h|r',
		data.mapId,
		data.level,
		data.mapNamePlain or data.mapName,
		data.level
	);

	return link;
end

function KeystoneManager:GetKeysText()
	local text = '';
	for char, key in pairs(self.db.keystones) do
		if 	key.level >= self.db.minLevel and
			key.level <= self.db.maxLevel and
			not self.db.excludes[key.mapId]
		then
			text = text .. self:NameWithoutRealm(char) .. ' - ' .. key.mapName .. ' +' .. key.level .. '\n';
		end
	end

	return text;
end

function KeystoneManager:ShowCopyWindow()
	if not self.KeystoneCopyWindow then
		local window = StdUi:Window(UIParent, 'Copy Keystones', 400, 350);
		window:SetPoint('CENTER');
		window:SetFrameStrata('FULLSCREEN_DIALOG');

		local mb = StdUi:MultiLineBox(window, 380, 300, self:GetKeysText());
		StdUi:GlueTop(mb.panel, window, 0, -40, 'CENTER');
		window.editBox = mb;
		self.KeystoneCopyWindow = window;
		return;
	end

	self.KeystoneCopyWindow:Show();
	self.KeystoneCopyWindow.editBox:SetText(self:GetKeysText());
end

function KeystoneManager:GetCurrentKeystoneInfo()
	local name = KeystoneManager:NameAndRealm();
	local keystone = KeystoneManager.db.keystones[name];

	return keystone;
end

function KeystoneManager:GetKeystone(force)
	force = force or false;
	
	if not self.db.keystones then
		self.db.keystones = {};
	end

	local name = self:NameAndRealm();
	local keystone = self.db.keystones[name];

	local mapId = C_MythicPlus.GetOwnedKeystoneChallengeMapID();
	local level = C_MythicPlus.GetOwnedKeystoneLevel();
	local mapName = self.mapNames[mapId];

	if not mapId or not level then
		return nil;
	end

	if force or not keystone or (level ~= keystone.level or mapId ~= keystone.mapId) then
		--keystone has changed
		if self.db.announce and not force then
			SendChatMessage('Keystone Manager: ' .. mapName .. ' +' .. level, 'PARTY');
		end

		self.db.keystones[name] = {
			mapId = mapId,
			mapName = mapName,
			level = level
		};

		self:UpdateTable(self.ScrollTable);
		self:RefreshDataText();
	end

	return keystone;
end

function KeystoneManager:UpdateWeeklyBest()
	local name = self:NameAndRealm();
	
	if not self.db.weeklyBest then
		self.db.weeklyBest = {};
	end

	local best = 0;
	for mapId, mapName in pairs(self.mapNames) do
		local _, level, _, _, members = C_MythicPlus.GetWeeklyBestForMap(mapId);
		
		if not level then
			level = 0;
		end

		if level > best then
			best = level;
		end
	end
	
	self.db.weeklyBest[name] = best;
	
	return best;
end

function KeystoneManager:PrintKeystone()
	local name = self:NameAndRealm();
	local keystone = self.db.keystones[name];
	if keystone then
		keystone = self:GetKeystone();
	end
	self:Print(self:CreateLink(keystone));
end

function KeystoneManager:ReportKeys()
	local target = self.db.whisper;
	if self.db.target ~= 'WHISPER' then
		target = nil;
	end

	for char, key in pairs(self.db.keystones) do
		
		if  key.level >= self.db.minLevel and
			key.level <= self.db.maxLevel and
			not self.db.excludes[key.mapId]
		then
			SendChatMessage(
				self:NameWithoutRealm(char) .. ' - ' .. key.mapName .. ' +' .. key.level,
				self.db.target,
				nil,
				target
			);
		end
	end
end

function KeystoneManager:ClearKeystones()
	self.db.weeklyBest = {};
	self.db.keystones = {};
	self:UpdateTable(self.ScrollTable);
	self:RefreshDataText()
end

function KeystoneManager:CreateGroup(info)
	--local groupName = '+' .. info.level;

	local activity = activities[info.mapId];

	PVEFrame_ShowFrame('GroupFinderFrame', LFGListPVEStub);

	local panel = LFGListFrame.CategorySelection;
	panel.selectedCategory = 2;

	local baseFilters = panel:GetParent().baseFilters;
	local entryCreation = panel:GetParent().EntryCreation;

	LFGListEntryCreation_Show(
		entryCreation, baseFilters, panel.selectedCategory, panel.selectedFilters
	);
	LFGListEntryCreation_Select(LFGListFrame.EntryCreation, nil, nil, nil, activity);

	--activity = 459;
	----activityID, itemLevel, honorLevel, autoAccept, privateGroup, questID
	--C_LFGList.CreateListing(activity, 0, 0, false, false, nil);
end

-- Helpers

function KeystoneManager:UpdateTable()
	local table = self.ScrollTable;
	if not table then
		return;
	end

	local tableData = {};
	for char, key in pairs(self.db.keystones) do
		
		local weeklyBest = self.db.weeklyBest[char];
		local color = self:GetKeystoneColor(key);

		tinsert(tableData, {
			name = self:NameWithoutRealm(char),
			weeklyBest = weeklyBest,
			mapName = format('|c%s%s|r', color, key.mapName),
			mapNamePlain = key.mapName,
			level = key.level,
			mapId = key.mapId
		});
	end

	table:SetData(tableData, true);
end

function KeystoneManager:RefreshDataText()
	local info = self:GetCurrentKeystoneInfo();
	if info then
		kmldbObject.text = self:GetShortInfo(info);
	else
		kmldbObject.text = 'Keystone Manager';
	end
end

function KeystoneManager:NameAndRealm()
	return UnitName('player') .. '-' .. GetRealmName();
end

function KeystoneManager:NameWithoutRealm(name)
	return gsub(name or '', '%-[^|]+', '');
end

function KeystoneManager:FormatKeystone(info)
	local color = self:GetKeystoneColor(info);
	return format('|c%s%s|r +%d', color, info.mapName, info.level);
end

function KeystoneManager:GetShortInfo(info)
	local color = self:GetKeystoneColor(info);
	return format('|c%s%s|r +%d', color, shortNames[info.mapId] or 'UNKNOWN', info.level);
end

function KeystoneManager:GetKeystoneColor(info)
	if self.db.excludes[info.mapId] then
		return 'ffc41f3b';
	end

	local quality = 2;

	local _, _, _, color = GetItemQualityColor(quality);

	return color;
end